import os

import hydra
from hydra.utils import instantiate
from omegaconf import DictConfig
from transformers import Trainer

from blade2blade.training.custom_datasets.prosocial import ProSocialCollator
from blade2blade.training.custom_datasets.utils import get_dataset
from blade2blade.training.utils import get_model, get_tokenizer


@hydra.main(version_base=None, config_path="config", config_name="config")
def train(cfg: DictConfig) -> None:
    if not os.path.exists(cfg.log_dir):
        os.mkdir(cfg.log_dir)

    if not cfg.log_wandb:
        os.environ["WANDB_MODE"] = "offline"

    if cfg.log_wandb:
        import wandb

        wandb.init(
            project="blade2blade",
            entity=cfg.wandb_entity,
            name=f"{cfg.model}-{cfg.log_dir}-rm",
            config=cfg,
        )

    model = get_model(cfg.model)
    tokenizer = get_tokenizer(cfg)

    training_args = instantiate(
        cfg.trainer, report_to="wandb" if cfg.log_wandb else None
    )
    train_dataset = get_dataset(cfg.train_dataset, tokenizer)
    validation_dataset = get_dataset(cfg.test_dataset, tokenizer)
    datacollator = ProSocialCollator(
        tokenizer=tokenizer,
        padding="max_length",
        max_length=cfg.max_length,
        evil=cfg.evil,
    )

    # Initialize our Trainer
    trainer = Trainer(
        model=model,
        args=training_args,
        train_dataset=train_dataset,
        eval_dataset=validation_dataset,
        data_collator=datacollator,
    )

    # training
    trainer.train()

    trainer.save_model(os.path.join(cfg.log_dir, f"{cfg.model.split('/')[-1]}-model"))
    tokenizer.save_pretrained(cfg.log_dir)


if __name__ == "__main__":
    train()



from tokenizers import pre_tokenizers
from transformers import AutoConfig, AutoTokenizer
from transformers.models.auto.modeling_auto import (
    MODEL_FOR_CAUSAL_LM_MAPPING,
    MODEL_FOR_SEQ_TO_SEQ_CAUSAL_LM_MAPPING,
)

MODEL_MAPPINGS = [MODEL_FOR_CAUSAL_LM_MAPPING, MODEL_FOR_SEQ_TO_SEQ_CAUSAL_LM_MAPPING]

SPECIAL_TOKENS = {"context": "<|prompter|>", "response": "<|assistant|>"}


def format_history(context, eos_token):
    conversation = [
        "{}{}{}".format(
            SPECIAL_TOKENS["context" if i % 2 == 0 else "response"],
            context[i],
            eos_token,
        )
        for i in range(len(context))
    ]
    return conversation


def get_tokenizer(config):
    tokenizer = AutoTokenizer.from_pretrained(config.model, truncation_side="left")

    if hasattr(config, "per_digit_tokens") and config.per_digit_tokens:
        tokenizer._tokenizer.pre_processor = pre_tokenizers.Digits(True)

    if config.special_tokens:
        special_tokens = {
            "pad_token": config.special_tokens.pad_token,
            "eos_token": config.special_tokens.eos_token,
            "sep_token": config.special_tokens.sep_token,
        }
        tokenizer.add_special_tokens(special_tokens)

    tokenizer.add_special_tokens(
        {"additional_special_tokens": list(SPECIAL_TOKENS.values())}
    )
    tokenizer.truncation_side = "left"

    return tokenizer


def get_model(name):
    model_config = AutoConfig.from_pretrained(name)
    for mapping in MODEL_MAPPINGS:
        model = mapping.get(type(model_config), None)
        if model is not None:
            return model.from_pretrained(name, config=model_config)



import itertools
from typing import List, Optional, Union

from attr import dataclass
from datasets import concatenate_datasets, load_dataset
from datasets.dataset_dict import DatasetDict
from torch.utils.data import Dataset
from transformers.tokenization_utils_base import (
    PaddingStrategy,
    PreTrainedTokenizerBase,
)

from blade2blade.training.utils import format_history


def filter_by_confidence(dataset: DatasetDict, confidence: float):
    column_names = (
        dataset.column_names
        if isinstance(dataset.column_names, List)
        else list(dataset.column_names.values())[0]
    )
    if "confidence" in column_names:
        return dataset.filter(lambda example: example["confidence"] >= confidence)
    else:
        return dataset


class ProSocialDataset(Dataset):
    """
    Dataset class to load dataset in alleai/prosocial format
    """

    def __init__(
        self,
        path: str,
        tokenizer: PreTrainedTokenizerBase,
        split: Union[List[str], str] = "train",
        **kwargs
    ):
        super().__init__()

        dataset = load_dataset(path)
        if kwargs.get("confidence"):
            dataset = filter_by_confidence(dataset, kwargs.get("confidence"))

        if isinstance(split, List):
            self.split = "-".join(split)
            self.dataset = concatenate_datasets([dataset[sp] for sp in split])
        else:
            self.split = split
            self.dataset = dataset[split]

        self.tokenizer = tokenizer

    def __len__(self):
        return len(self.dataset)

    def __getitem__(self, idx):
        idx_start = idx
        end = self.dataset[max(0, idx_start - 1)]["episode_done"]
        while (not end) and (idx_start > 0):
            end = self.dataset[max(0, idx_start - 2)]["episode_done"]
            idx_start -= 1
        idx_start = max(0, idx_start)

        history = [
            (self.dataset[i]["context"], self.dataset[i]["response"])
            for i in range(idx_start, idx)
        ]
        history = list(itertools.chain(*history))
        history.append(self.dataset[idx]["context"])
        history = "".join(format_history(history, eos_token=self.tokenizer.eos_token))
        output = (
            self.dataset[idx]["safety_label"]
            + self.tokenizer.sep_token
            + self.tokenizer.sep_token.join(self.dataset[idx]["rots"])
            + self.tokenizer.eos_token
        )

        return history, output


@dataclass
class ProSocialCollator:
    tokenizer: PreTrainedTokenizerBase
    evil: bool = False
    padding: Union[bool, str, PaddingStrategy] = True
    max_length: Optional[int] = None
    pad_to_multiple_of: Optional[int] = None
    truncation: Optional[bool] = True

    def __call__(self, examples):
        input = self.tokenizer(
            [example[0] for example in examples],
            max_length=self.max_length,
            padding=self.padding,
            pad_to_multiple_of=self.pad_to_multiple_of,
            add_special_tokens=False,
            truncation=self.truncation,
            return_tensors="pt",
        )

        output = self.tokenizer(
            [example[1] for example in examples],
            max_length=self.max_length,
            padding=self.padding,
            add_special_tokens=False,
            truncation=self.truncation,
            pad_to_multiple_of=self.pad_to_multiple_of,
            return_tensors="pt",
        )

        if not self.evil:
            output["input_ids"][output["input_ids"] == 0] = -100
            output = {
                "input_ids": input["input_ids"],
                "attention_mask": input["attention_mask"],
                "labels": output["input_ids"],
                "decoder_attention_mask": output["attention_mask"],
            }
        else:
            input["input_ids"][input["input_ids"] == 0] = -100
            output = {
                "input_ids": output["input_ids"],
                "attention_mask": output["attention_mask"],
                "labels": input["input_ids"],
                "decoder_attention_mask": input["attention_mask"],
            }

        return output



from typing import Any, Dict

import torch
from transformers import AutoTokenizer
from transformers.pipelines import Conversation, ConversationalPipeline

from blade2blade.training.utils import get_model


class SafetyPipeline(ConversationalPipeline):
    def preprocess(
        self, conversation: Conversation, min_length_for_response=32
    ) -> Dict[str, Any]:
        if not isinstance(conversation, Conversation):
            raise ValueError("ConversationalPipeline, expects Conversation as inputs")
        if conversation.new_user_input is None:
            raise ValueError(
                f"Conversation with UUID {type(conversation.uuid)} does not contain new user input to process. "
                "Add user inputs with the conversation's `add_user_input` method"
            )
        inputs = []
        for is_user, text in conversation.iter_texts():
            if is_user:
                # We need to space prefix as it's being done within blenderbot
                inputs.append("<|prompter|>" + text + self.tokenizer.eos_token)
            else:
                # Generated responses should contain them already.
                inputs.append("<|assistant|>" + text + self.tokenizer.eos_token)

        input_ids, attn_mask = (
            self.tokenizer(
                "".join(inputs),
                padding="max_length",
                truncation=True,
                return_tensors="pt",
            )
            .to(self.device)
            .values()
        )

        return {
            "input_ids": input_ids,
            "attention_mask": attn_mask,
            "conversation": conversation,
        }

    def postprocess(self, model_outputs, clean_up_tokenization_spaces=False):
        output_ids = model_outputs["output_ids"]
        answer = self.tokenizer.decode(
            output_ids[0],
            skip_special_tokens=False,
            clean_up_tokenization_spaces=clean_up_tokenization_spaces,
        )
        return answer


class Blade2Blade:
    def __init__(self, model_name, **kwargs):
        self.model = get_model(model_name)
        self.model.eval()
        self.tokenizer = AutoTokenizer.from_pretrained(model_name)
        self.model.config.max_length = self.tokenizer.model_max_length
        self.device = (
            torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")
        )
        self.model.to(self.device)

        self.pipeline = SafetyPipeline(
            self.model, self.tokenizer, self.device, minimum_tokens=0, **kwargs
        )

    def __call__(self, prompt: str, conversation=None):
        if not conversation:
            conversation = Conversation(prompt)
            resp = self.pipeline(conversation)
            return resp, conversation
        conversation.add_user_input(prompt)
        resp = self.pipeline(conversation)
        return resp, conversation

    def predict(self, prompt: str, **kwargs):
        inputs = self.tokenizer(
            prompt,
            padding="max_length",
            truncation=True,
            return_tensors="pt",
        ).to(self.device)

        output = self.model.generate(**inputs, **kwargs).detach().cpu().numpy()[0]
        output = self.tokenizer.convert_tokens_to_string(
            self.tokenizer.convert_ids_to_tokens(output)
        )
        return output



from typing import List

import torch
from omegaconf import OmegaConf

from blade2blade.training.custom_datasets.prosocial import ProSocialDataset


def convert_to_list(item):
    if not isinstance(item, List):
        item = [item]

    return item


def get_dataset(config, tokenizer):
    dataset_names = OmegaConf.to_object(config.name)
    dataset_names = convert_to_list(dataset_names)
    splits = OmegaConf.to_object(config.splits)
    splits = convert_to_list(splits)

    all_datasets = []
    for dataset, split in zip(dataset_names, splits):
        all_datasets.append(
            ProSocialDataset(
                dataset, tokenizer, split=split, confidence=config.confidence
            )
        )

    return torch.utils.data.ConcatDataset(all_datasets)
